import typer

from chitra import __version__
from chitra.cli import builder

app = typer.Typer(
    name="chitra CLI ✨",
    add_completion=False,
)

app.add_typer(
    builder.app,
    name="builder",
)


@app.command()
def version():
    typer.echo(f"Hey 👋! You're running chitra version={__version__} ✨")



import asyncio
from unittest.mock import MagicMock

import numpy as np
import pytest
from fastapi import FastAPI

from chitra.serve import API
from chitra.serve import constants as const
from chitra.serve import create_api


def dummy_model(x):
    return np.asarray([1, 2])


def test_create_app():
    api_types = API.get_available_api_types()

    for api_type in api_types:
        api = create_api(dummy_model, api_type)
        assert isinstance(api.app, FastAPI)


def test_api():
    api_types = API.get_available_api_types()
    for api_type in api_types:
        api = API(api_type, dummy_model)
        assert isinstance(api.app, FastAPI)


@pytest.mark.asyncio
async def test_predict_image():
    def async_return(result):
        f = asyncio.Future()
        f.set_result(result)
        return f

    app = API(api_type=const.IMAGE_CLF, model=dummy_model)
    app.data_processor = MagicMock()
    app.data_processor.preprocess_fn = MagicMock()
    app.data_processor.postprocess_fn = lambda x: x
    file = MagicMock()
    file.read = MagicMock(return_value=async_return("Sample string"))
    output = await app.predict_image(file)
    assert isinstance(output, np.ndarray)
    assert len(output) == 2



import tensorflow as tf

from chitra.core import load_imagenet_labels
from chitra.serve.app import GradioApp

image_shape = (224, 224)
model = tf.keras.applications.MobileNetV2(weights="imagenet")
IMAGENET_LABELS = load_imagenet_labels()


def postprocess(preds):
    preds = tf.argmax(preds, 1).numpy()
    label = IMAGENET_LABELS[preds[0]]
    return label


app = GradioApp(
    "image-classification",
    model=model,
    image_shape=image_shape,
    postprocess_fn=postprocess,
)

app.run(share=True)



import matplotlib.pyplot as plt
import rich

from chitra import Chitra

url = "https://upload.wikimedia.org/wikipedia/commons/thumb/b/b6/Image_created_with_a_mobile_phone.png/1200px-Image_created_with_a_mobile_phone.png"
box = [[600, 250, 900, 600.1]]
label = ["handphone"]

image = Chitra(url, box, label)
rich.print("before resize:", image.bboxes)

image.resize_image_with_bbox((224, 224))
rich.print("after resize:", image.bboxes)

plt.imshow(image.draw_boxes())



import numpy as np
from imgaug.augmentables import bbs

from chitra.coordinates import BoundingBoxes


def test_bounding_boxes():
    box = [1, 2, 3, 4]
    label = ["Dog"]
    bounding_box = BoundingBoxes(box, label)
    assert len(bounding_box.bboxes) == 1
    assert isinstance(bounding_box.bboxes, list)
    assert isinstance(bounding_box.bboxes[0], bbs.BoundingBox)


def test_corner_to_center():
    xmin, ymin, xmax, ymax = 0, 0, 10, 10
    cx, cy, h, w = BoundingBoxes.corner_to_center(xmin, ymin, xmax, ymax)
    assert (cx, cy, h, w) == (5, 5, 10, 10)


def test_resize_with_image():
    box = [1, 2, 3, 4]
    label = ["chitra"]
    bounding_box = BoundingBoxes(box, label)
    bbs = bounding_box.resize_with_image((10, 10, 3), np.random.randn(100, 100, 3))
    rescaled_bounding_box = bbs.bounding_boxes[0]
    assert np.isclose(rescaled_bounding_box.x1, 10)
    assert np.isclose(rescaled_bounding_box.y1, 20)
    assert np.isclose(rescaled_bounding_box.x2, 30)
    assert np.isclose(rescaled_bounding_box.y2, 40)



import inspect
from functools import partial
from typing import Any, List, Optional, Union

import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
import tensorflow_addons as tfa
from PIL import Image
from tensorflow import keras
from tensorflow.keras.models import Model
from tf_keras_vis.gradcam import Gradcam, GradcamPlusPlus
from tf_keras_vis.utils import normalize
from typeguard import check_argument_types, typechecked

from chitra.import_utils import is_installed

from .converter.core import pytorch_to_onnx, tf2_to_onnx
from .datagenerator import Dataset

pl = None
if is_installed("pytorch_lightning"):
    import pytorch_lightning as pl

MODEL_DICT = {}
for name, func in inspect.getmembers(tf.keras.applications):
    if inspect.isfunction(func):
        MODEL_DICT[name.lower()] = func

OPT_DICT = {}
for name, func in inspect.getmembers(tf.keras.optimizers):
    if inspect.isclass(func):
        OPT_DICT[name.lower()] = func


@typechecked
def _get_base_cnn(
    base_model: Union[str, Model],
    pooling: str = "avg",
    weights: Union[str, None] = "imagenet",
    include_top: bool = False,
) -> Model:
    if isinstance(base_model, str):
        if base_model not in MODEL_DICT.keys():
            raise AssertionError(
                f"base_model name must be in {tuple(MODEL_DICT.keys())}"
            )
        base_model = MODEL_DICT[base_model]
        base_model = base_model(
            include_top=include_top, pooling=pooling, weights=weights
        )
    return base_model


@typechecked
def _add_output_layers(
    base_model: Model,
    outputs: int,
    drop_out: Union[float, None] = None,
    name: Optional[str] = None,
) -> Model:
    x = base_model.output
    if drop_out:
        x = tf.keras.layers.Dropout(drop_out)(x)
    x = tf.keras.layers.Dense(outputs, name="output")(x)

    model = tf.keras.Model(base_model.input, x, name=name)
    return model


def create_classifier(
    base_model_fn: callable,
    num_classes: int,
    weights="imagenet",
    dropout: Optional[float] = None,
    include_top=False,
):
    base_model = base_model_fn(
        include_top=include_top,
        weights=weights,
    )
    if include_top:
        return base_model

    outputs = 1 if num_classes == 2 else num_classes

    x = base_model.output
    x = tf.keras.layers.GlobalMaxPool2D()(x)
    if dropout:
        x = tf.keras.layers.Dropout(dropout)(x)
    x = tf.keras.layers.Dense(outputs, name="output")(x)

    model = tf.keras.Model(base_model.input, x)
    return model


@typechecked
def create_cnn(
    base_model: Union[str, Model],
    num_classes: int,
    drop_out=0.5,
    keras_applications: bool = True,
    pooling: str = "avg",
    weights: Union[str, None] = "imagenet",
    name: Optional[str] = None,
) -> Model:
    assert pooling in ("avg", "max")

    if keras_applications:
        if num_classes == 2:
            outputs = 1
        else:
            outputs = num_classes
    else:
        print("num_classes is ignored. returning the passed model as it is.")

    if isinstance(base_model, (str, Model)) and keras_applications:
        base_model = _get_base_cnn(base_model, pooling=pooling, weights=weights)
        if "pool" not in base_model.layers[-1].name:
            raise AssertionError("base_model last layer must be a pooling layer")
        model = _add_output_layers(base_model, outputs, drop_out=drop_out, name=name)

    elif isinstance(base_model, Model) and keras_applications is False:
        model = base_model

    elif isinstance(base_model, str) and keras_applications is False:
        model = _get_base_cnn(base_model, weights="imagenet", include_top=True)

    else:
        print("Invalid arguments!")
    return model


class Trainer(Model):
    """The Trainer class inherits tf.keras.Model and contains everything a
    model needs for training. It exposes trainer.cyclic_fit method which trains
    the model using Cyclic Learning rate discovered by Leslie Smith.

    Arguments:
    ds: Dataset object
    model: object of type tf.keras.Model
    num_classes (int, None): number of classes in the dataset. If None then will auto infer from Dataset
    """

    _AUTOTUNE = tf.data.experimental.AUTOTUNE

    @typechecked
    def __init__(
        self, ds: Dataset, model: Model, num_classes: Union[int, None] = None, **kwargs
    ):
        assert check_argument_types()

        super(Trainer, self).__init__()
        self.ds = ds
        self.total = len(ds)
        if num_classes:
            self.NUM_CLASSES = num_classes
        else:
            self.NUM_CLASSES = ds.NUM_CLASSES
        self.gradcam = None
        self.model = model
        self.cyclic_opt_set = False
        self.max_lr, self.min_lr = None, None
        self.batch_size = None
        self.step_size = None

    def build(self):
        raise NotImplementedError(
            'Build method is not implemented in Trainer! Please use "model.model.build" instead.'
        )

    def summary(self):
        return self.model.summary()

    def compile(self, *args, **kwargs):
        return self.model.compile(*args, **kwargs)

    def call(self, *args, **kwargs):
        return self.model.call(*args, **kwargs)

    def fit(self, *args, **kwargs):
        return self.model.fit(*args, **kwargs)

    def warmup(self):
        raise NotImplementedError(
            "warmup is not implemented yet! Would you like to raise a PR to chitra?"
        )

    @staticmethod
    def prewhiten(image):
        image = tf.cast(image, tf.float32)
        image = image / 127.5 - 1.0
        return image

    def rescale(self, image, label):
        image = self.prewhiten(image)
        return image, label

    def _get_optimizer(self, optimizer, momentum=0.9, **kwargs):
        if optimizer.__name__ == "SGD":
            optimizer = partial(
                optimizer, momentum=momentum, nesterov=kwargs.get("nesterov", True)
            )
        else:
            optimizer = partial(
                optimizer,
            )
        return optimizer

    def _prepare_dl(self, bs: int = 8, shuffle: bool = True):
        ds = self.ds
        dl = ds.get_tf_dataset(shuffle=shuffle)
        dl = dl.map(self.rescale, Trainer._AUTOTUNE)
        return dl.batch(bs).prefetch(Trainer._AUTOTUNE)

    def cyclic_fit(
        self,
        epochs: int,
        batch_size: int,
        lr_range: Union[tuple, list] = (1e-4, 1e-2),
        optimizer: tf.keras.optimizers.Optimizer = tf.keras.optimizers.SGD,
        momentum: float = 0.9,
        validation_data: Any = None,
        callbacks: Optional[List] = None,
        *args,
        **kwargs,
    ):
        """Trains model on ds as train data with cyclic learning rate. Dataset
        will be automatically converted into `tf.data` format and images will
        be prewhitened in range of [-1, 1]. Cyclical Learning Rates for
        Training Neural Networks: https://arxiv.org/abs/1506.01186.

        Args:
            epochs (int): number of epochs for training
            batch_size (int): batch size
            lr_range (tuple): learning rate will cycle from lr_min to lr_max
            optimizer (callable): Keras callable optimizer
            momentum(float): momentum for the optimizer
            validation_data: Data on which to evaluate
            callbacks: List of `tf.keras.callbacks` instances.
        kwargs:
            step_size (int): step size for the Cyclic learning rate. By default it is `2 * len(self.ds)//batch_size`
            scale_mode (str): cycle or exp
            shuffle(bool): Dataset will be shuffle on each epoch if True
        """
        if not self.cyclic_opt_set:
            self.max_lr, self.min_lr = lr_range
            step_size = 2 * len(self.ds) // batch_size
            lr_schedule = tfa.optimizers.Triangular2CyclicalLearningRate(
                initial_learning_rate=lr_range[0],
                maximal_learning_rate=lr_range[1],
                step_size=kwargs.get("step_size", step_size),
                scale_mode=kwargs.get("scale_mode", "cycle"),
            )

            optimizer = self._get_optimizer(optimizer, momentum=momentum)
            optimizer = optimizer(learning_rate=lr_schedule)
            self.model.optimizer = optimizer
            self.cyclic_opt_set = True
        else:
            print("cyclic learning rate already set!")

        return self.model.fit(
            self._prepare_dl(batch_size, kwargs.get("shuffle", True)),
            validation_data=validation_data,
            epochs=epochs,
            callbacks=callbacks,
        )

    @typechecked
    def compile2(
        self,
        batch_size: int,
        optimizer: Union[str, tf.keras.optimizers.Optimizer] = "adam",
        lr_range: Union[tuple, list] = (1e-4, 1e-2),
        loss: Optional[tf.keras.losses.Loss] = None,
        metrics=None,
        **kwargs,
    ):
        """Compile2 compiles the model of Trainer for cyclic learning rate.
        Cyclical Learning Rates for Training Neural Networks:
        https://arxiv.org/abs/1506.01186.

        Args:
            batch_size (int): batch size
            lr_range (tuple): learning rate will cycle from lr_min to lr_max
            optimizer (str, keras.optimizer.Optimizer): Keras optimizer

        kwargs:
            step_size (int): step size for the Cyclic learning rate. By default it is `2 * len(self.ds)//batch_size`
            scale_mode (str): cycle or exp
            momentum(int): momentum for the optimizer when optimizer is of type str
        """
        self.max_lr, self.min_lr = lr_range
        self.batch_size = batch_size

        self.step_size = step_size = 2 * len(self.ds) // batch_size

        lr_schedule = tfa.optimizers.Triangular2CyclicalLearningRate(
            initial_learning_rate=lr_range[0],
            maximal_learning_rate=lr_range[1],
            step_size=kwargs.get("step_size", step_size),
            scale_mode=kwargs.get("scale_mode", "cycle"),
        )

        if isinstance(optimizer, str):
            optimizer = OPT_DICT[optimizer]
            optimizer = optimizer(learning_rate=lr_schedule)
            if kwargs.get("momentum"):
                optimizer.momentum = kwargs.get("momentum")

        else:
            optimizer.learning_rate = lr_schedule

        self.compile(optimizer=optimizer, loss=loss, metrics=metrics)
        self.cyclic_opt_set = True
        print("Model compiled!")


class InterpretModel:
    def __init__(self, gradcam_pp: bool, learner: Trainer, clone: bool = False):
        """Args:
        gradcam_pp: if True GradCam class will be used else GradCamPlusplus
        clone: whether GradCam will clone learner.model
        """
        if gradcam_pp:
            self.gradcam_fn = GradcamPlusPlus
        else:
            self.gradcam_fn = Gradcam
        self.learner = learner

        self.gradcam = self.gradcam_fn(learner.model, self.model_modifier, clone=clone)

    def __call__(
        self, image: Image.Image, auto_resize: bool = True, image_size=None
    ) -> None:
        gradcam = self.gradcam
        get_loss = self.get_loss
        if auto_resize and image_size is None:
            image_size = self.learner.ds.img_sz_list.get_size()
        if image_size:
            image = image.resize(image_size)

        X = np.asarray(image, np.float32)
        X = self.learner.prewhiten(X)
        X = np.expand_dims(X, 0)

        cam = gradcam(
            get_loss,
            X,
            penultimate_layer=-1,  # model.layers number
            seek_penultimate_conv_layer=True,
        )
        cam = normalize(cam)
        heatmap = np.uint8(cm.jet(cam[0])[..., :3] * 255)
        plt.imshow(image)
        plt.imshow(heatmap, cmap="jet", alpha=0.5)
        plt.show()

    @staticmethod
    def model_modifier(m):
        """Sets last activation to linear."""
        m.layers[-1].activation = tf.keras.activations.linear
        return m

    def get_loss(self, preds):
        if self.learner.NUM_CLASSES == 2:
            ret = preds[0]
        else:
            index = tf.argmax(tf.math.softmax(preds), axis=1)[0]
            ret = preds[0, index]
            print(f"index: {index}")
        return ret


class Learner:
    TF = ("TF", "TENSORFLOW")
    PT = ("PYTORCH", "PT", "TORCH")

    def __init__(
        self, model: Union["pl.LightningModule", "keras.models.Model"], mode: str = "TF"
    ):
        self.MODE = mode.upper()
        self.model = model
        self.epochs_trained = 0

        if self.MODE in Learner.PT:
            self.trainer = None

    def fit(
        self,
        train_data,
        epochs,
        val_data=None,
        test_data=None,
        callbacks=None,
        **kwargs,
    ):
        """train models For TF:

        Just pass train data and start training For PyTorch:     You can
        enter configs to Lightning Trainer
        """
        MODE = self.MODE
        initial_epoch = self.epochs_trained
        self.epochs_trained += epochs

        if MODE in Learner.TF:
            return self.model.fit(
                train_data,
                epochs=epochs,
                initial_epoch=initial_epoch,
                validation_data=val_data,
                callbacks=callbacks,
            )
        if MODE in Learner.PT:
            lit_confs = kwargs.get("LIT_TRAINER_CONFIG", {})
            if not self.trainer:
                self.trainer = pl.Trainer(max_epochs=epochs, **lit_confs)
            return self.trainer.fit(self.model, train_data, val_data)

    def to_onnx(self, tensor=None, export_path=None):
        MODE = self.MODE
        if MODE in Learner.TF:
            return tf2_to_onnx(self.model, output_path=export_path)

        if MODE in Learner.PT:
            return pytorch_to_onnx(self.model, tensor, export_path)



from unittest.mock import MagicMock

import numpy as np
from PIL import Image

from chitra.image import Chitra, _cache_image

url = (
    "https://raw.githubusercontent.com/aniketmaurya/chitra/master/docs/assets/logo.png"
)
image = Chitra(url, cache=True)


def test__load_image():
    url = "https://raw.githubusercontent.com/aniketmaurya/chitra/master/docs/assets/logo.png"
    image = Chitra(url, cache=True)
    assert isinstance(image.image, Image.Image)


def test_numpy():
    assert isinstance(image.numpy(), np.ndarray)


def test_to_tensor():
    assert True


def test_shape():
    assert len(image.shape) == 3


def test_size():
    assert len(image.size) == 2


def test_imshow():
    assert True


def test_draw_boxes():
    assert True


def test_resize_image_with_bbox():
    box = [10, 20, 30, 40]
    label = ["chitra"]
    dummy = np.random.randn(100, 100, 3).astype("uint8")
    image = Chitra(dummy, bboxes=box, labels=label)
    image.resize_image_with_bbox((10, 10))
    rescaled_bounding_box = image.bboxes[0]

    assert np.isclose(rescaled_bounding_box.x1, 1)
    assert np.isclose(rescaled_bounding_box.y1, 2)
    assert np.isclose(rescaled_bounding_box.x2, 3)
    assert np.isclose(rescaled_bounding_box.y2, 4)


def test__cache_image():
    image = MagicMock()
    image.save = MagicMock()
    _cache_image(image, "test_image.jpg")
    image.save.assert_called_once()



from typing import Any, Callable, Optional

import grpc
import tensorflow as tf
from tensorflow_serving.apis import predict_pb2, prediction_service_pb2_grpc


def create_grpc_stub(
    host: str = "localhost", port: str = "8500"
) -> prediction_service_pb2_grpc.PredictionServiceStub:
    hostport = f"{host}:{port}"
    channel = grpc.insecure_channel(hostport)
    stub = prediction_service_pb2_grpc.PredictionServiceStub(channel)
    return stub


def grpc_request(
    stub: prediction_service_pb2_grpc.PredictionServiceStub,
    data_sample: Any,
    input_name: str,
    model_name: str,
    signature_name: str,
    callback: Optional = None,
    grpc_timeout: int = 20,
    async_: bool = False,
):
    request = predict_pb2.PredictRequest()
    request.model_spec.name = model_name
    request.model_spec.signature_name = signature_name

    request.inputs[input_name].CopyFrom(
        tf.make_tensor_proto(data_sample, shape=data_sample.shape)
    )

    if async_:
        result_future = stub.Predict.future(request, 5)  # 5 seconds
    else:
        result_future = stub.Predict(request, grpc_timeout)

    if callback is not None:
        return callback(result_future)

    return result_future


class GrpcClient:
    def __init__(self, host: str = "localhost", port: str = "8500"):
        super().__init__()
        self.stub = create_grpc_stub(host, port)

    def request(
        self,
        data_sample,
        input_name: str,
        model_name: str,
        signature_name: str,
        callback: Callable = None,
        grpc_timeout: int = 20,
        async_: bool = False,
    ):
        stub = self.stub
        response = grpc_request(
            stub,
            data_sample,
            input_name,
            model_name,
            signature_name,
            callback,
            grpc_timeout,
            async_,
        )
        return response



import pytest

from chitra.data_processing import DataProcessor


def dummy_preprocess(x: int):
    return x + 1


def dummy_postprocess(x: int):
    return x - 1


def test_data_processor():
    x = 5
    data_processor_empty = DataProcessor()
    assert not data_processor_empty._preprocess_fn
    assert not data_processor_empty._postprocess_fn

    with pytest.raises(UserWarning):
        data_processor_empty.preprocess(x)

    with pytest.raises(UserWarning):
        data_processor_empty.postprocess(x)

    data_processor = DataProcessor(
        preprocess_fn=dummy_preprocess, postprocess_fn=dummy_postprocess
    )

    x = data_processor.preprocess(x)
    x = data_processor.postprocess(x)
    assert x == 5

    data_processor_empty.set_preprocess_fn(dummy_preprocess)
    data_processor_empty.set_postprocess_fn(dummy_postprocess)

    x = data_processor_empty.preprocess(x)
    x = data_processor_empty.postprocess(x)
    assert x == 5



from io import BytesIO
from typing import List, Optional, Tuple

import numpy as np
from PIL import Image

from chitra.image import Chitra

from ..processor import DataProcessor


def default_preprocess(
    data,
    image_shape: Optional[Tuple[int, int]] = None,
    rescale: bool = True,
    expand_dims: bool = True,
) -> np.ndarray:
    """
    Supports image resize, rescaling and dimension expansion along 0th index
    Args:
        data: Image File buffer or numpy array
        image_shape: Target image size
        rescale: If true then image will be rescaled as [-1, 1]
        expand_dims:

    Returns:
        preprocessed numpy array image
    """
    if isinstance(data, str):
        image = Image.open(BytesIO(data)).convert("RGB")
    elif isinstance(data, np.ndarray):
        image = Chitra(data).image
    else:
        raise UserWarning(f"preprocessing not implemented for this data type -> {data}")

    if image_shape:
        image = image.resize(image_shape)

    image = np.asarray(image).astype(np.float32)
    if rescale:
        image = image / 127.5 - 1.0
    if expand_dims:
        image = np.expand_dims(image, 0)
    return image


def default_postprocess(data, return_type: Optional[str] = "list") -> List:
    if not isinstance(data, (np.ndarray, list, tuple, int, float)):
        data = data.numpy()
    if return_type == "list":
        if isinstance(data, np.ndarray):
            data = data.tolist()
        else:
            list(data)
    return data


class DefaultVisionProcessor:
    vision = DataProcessor(default_preprocess, default_postprocess)



from typing import Callable, List, Optional, Union

import gradio as gr
import numpy as np

from chitra.__about__ import documentation_url
from chitra.serve import constants as const
from chitra.serve.base import ModelServer


class GradioApp(ModelServer):
    API_TYPES = {"VISION": (const.IMAGE_CLF, const.OBJECT_DETECTION)}

    def __init__(
        self,
        api_type: str,
        model: Callable,
        input_types: Optional[Union[List, str]] = None,
        output_types: Optional[Union[List, str]] = None,
        preprocess_fn: Callable = None,
        postprocess_fn: Callable = None,
        preprocess_conf: Optional[dict] = None,
        postprocess_conf: Optional[dict] = None,
        **kwargs,
    ):
        super(GradioApp, self).__init__(
            api_type, model, preprocess_fn, postprocess_fn, **kwargs
        )
        if not preprocess_conf:
            preprocess_conf = {}
        if not postprocess_conf:
            postprocess_conf = {}

        self.title = kwargs.get("title", "Chitra Server")
        self.desc = kwargs.get(
            "description",
            f"<a href={documentation_url}>Docs</a> 🔗",
        )
        self.input_types = input_types
        self.output_types = output_types
        self.api_type_func = {}
        self.preprocess_conf = preprocess_conf
        self.postprocess_conf = postprocess_conf
        self.setup(**kwargs)

    def setup(
        self,
        **kwargs,
    ):
        if self.api_type in (const.IMAGE_CLF, const.OBJECT_DETECTION):
            self.api_type_func[self.api_type] = self.single_x_classification
        elif self.api_type == const.TXT_CLF:
            self.api_type_func[self.api_type] = self.single_x_classification
        else:
            raise NotImplementedError(f"api_type={self.api_type} not implemented yet!")

        if not self.input_types:
            self.input_types = self.get_input_type(**kwargs)

        if not self.output_types:
            self.output_types = "json"

    def get_input_type(self, **kwargs):
        label = kwargs.get("label")
        if self.api_type in (const.IMAGE_CLF, const.OBJECT_DETECTION):
            return gr.inputs.Image(shape=kwargs.get("image_shape"), label=label)

        if self.api_type == const.TXT_CLF:
            return gr.inputs.Textbox(
                lines=2, placeholder=kwargs.get("text_placeholder"), label=label
            )
        raise NotImplementedError(f"{self.api_type} API Type is not implemented yet!")

    def single_x_classification(self, x: np.ndarray):
        data_processor = self.data_processor

        if data_processor.preprocess_fn:
            x = data_processor.preprocess(x, **self.preprocess_conf)
        x = self.model(x)
        if data_processor.postprocess_fn:
            x = data_processor.postprocess(x, **self.postprocess_conf)
        return x

    def run(self, share: bool = False, gr_interface_conf: Optional[dict] = None):
        if not gr_interface_conf:
            gr_interface_conf = {}

        gr.Interface(
            fn=self.api_type_func[self.api_type],
            inputs=self.input_types,
            outputs=self.output_types,
            title=self.title,
            description=self.desc,
            **gr_interface_conf,
        ).launch(share=share)



import os
import shlex
import subprocess
from glob import glob
from pathlib import Path
from typing import List, Optional

import typer

import chitra

app = typer.Typer(
    help="""Auto Builds docker image for chitra ModelServer.
    path should contain a `main.py` file which will have an object of type `chitra.serve.ModelServer` and
    its name should be `app`. If you have any external Python dependency then create a `requirements.txt` file and
    keep in the same directory."""
)


def get_dockerfile() -> str:
    path = Path(os.path.dirname(chitra.__file__)) / "assets/API.Dockerfile"
    with open(path, "r") as fr:
        data = fr.read()
    return data


def file_check(files: List) -> None:
    files = map(os.path.basename, files)
    if "requirements.txt" not in files:
        raise UserWarning("requirements.txt not found!")

    if "main.py" not in files:
        raise UserWarning(
            "main.py not found! Your main.py should contain app \
            object of type chitra.serve.ModelServer"
        )


def text_to_file(text: str, path: str):
    with open(path, "w") as fw:
        fw.write(text)


@app.command()
def create(
    path: str = "./",
    port: Optional[str] = None,
    tag: Optional[str] = None,
):
    """
    Auto-builds Docker image for chitra ModelServer

    Args:

        path: file-location of main.py

        port: port to expose

        tag: tag of docker image

    """
    if not port:
        port = "8080"
    if not tag:
        tag = "chitra-server"

    path = Path(path)
    files = glob(str(path / "*"))
    file_check(files)
    style_path = typer.style(str(path), fg=typer.colors.GREEN)
    typer.echo(f"Everything under {style_path} will be added to Docker image!")
    show_files = typer.confirm("Do you wish to see the files to be added?")
    if show_files:
        typer.echo(files)

    dockerfile = get_dockerfile()
    dockerfile = dockerfile.replace("PORT", port)
    if typer.confirm("Show Dockerfile"):
        typer.echo(dockerfile)
    text_to_file(dockerfile, "Dockerfile")

    typer.echo(f"Building Docker {tag} 🐳")
    cmd = f"docker build --tag {tag} ."
    cmd = shlex.split(cmd)
    process = subprocess.run(cmd, stdout=subprocess.PIPE, universal_newlines=True)
    if process.returncode == 0:
        typer.secho(
            "Docker image has been created for your app! Check the image using ",
            nl=False,
            fg=typer.colors.GREEN,
        )
        typer.secho("docker images", fg=typer.colors.BRIGHT_BLUE)
    else:
        typer.secho("Docker build failed!", fg=typer.colors.RED)



from unittest.mock import MagicMock, patch

import numpy as np
from tensorflow_serving.apis.prediction_service_pb2_grpc import PredictionServiceStub

from chitra.serve.tf_serving.client import GrpcClient, create_grpc_stub, grpc_request


def test_create_grpc_stub():
    assert isinstance(create_grpc_stub(), PredictionServiceStub)


def test_request():
    client = GrpcClient()
    assert isinstance(client.stub, PredictionServiceStub)


@patch("chitra.serve.tf_serving.client.predict_pb2")
def test_grpc_request(mock_predict_pb2):
    mock_predict_pb2.PredictRequest = MagicMock()
    stub = MagicMock()
    stub.Predict = MagicMock()
    stub.Predict.return_value = True

    data = np.random.randn(224, 224, 3)

    result = grpc_request(
        stub,
        data,
        input_name="input",
        model_name="test_model",
        signature_name="test",
    )

    assert result
    stub.Predict.assert_called()
    mock_predict_pb2.PredictRequest.assert_called()
    stub.Predict.assert_called_with(mock_predict_pb2.PredictRequest(), 20)



from chitra.visualization.metrics import plot_confusion_matrix

y_pred = [1, 1, 0, 1]
y_true = [0, 1, 0, 1]
display_labels = ("class A", "class B")

plot_confusion_matrix(y_pred, y_true, display_labels=display_labels)



"""https://www.kaggle.com/grfiv4/plot-a-confusion-matrix."""
from itertools import product
from typing import List, Union

import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import confusion_matrix

from chitra.logging import logger


def detect_multilabel(labels: Union[List, np.ndarray]):
    if isinstance(labels, (np.ndarray, list, tuple)):
        labels = np.unique(labels)
    else:
        raise UserWarning(
            f"expect data type for label was list or np.ndarray but got {type(labels)}"
        )

    if len(labels) > 2:
        return True
    return False


def cm_accuracy(cm: np.ndarray):
    return np.trace(cm) / float(np.sum(cm))


def plot_confusion_matrix(
    y_pred: Union[np.ndarray, List],
    y_true: Union[np.ndarray, List],
    display_labels=None,
    include_values: bool = True,
    title: str = "Confusion Matrix",
    cmap: str = None,
):
    if detect_multilabel(y_true):
        logger.warning("You might want to use multi-label version!")

    if display_labels is None:
        display_labels = np.unique(y_true)

    n_classes = len(display_labels)
    tick_marks = np.arange(n_classes)

    if cmap is None:
        cmap = plt.get_cmap("Blues")

    cm = confusion_matrix(y_true, y_pred)
    accuracy = cm_accuracy(cm)
    error = 1 - accuracy

    plt.imshow(cm, cmap=cmap)

    if include_values:
        for i, j in product(range(n_classes), range(n_classes)):
            plt.text(i, j, "{:,}".format(cm[i, j]))

    plt.xticks(tick_marks, display_labels, rotation=45)
    plt.yticks(tick_marks, display_labels)
    plt.title(title)
    plt.xlabel(f"Predicted Label\nAccuracy={accuracy:0.4f}; Error={error:0.4f}")
    plt.ylabel("True Label")

    plt.show()



from unittest.mock import MagicMock, patch

import pytest
import typer
from typer.testing import CliRunner

from chitra.cli import builder
from chitra.cli.builder import file_check

runner = CliRunner()


@patch("chitra.cli.builder.subprocess")
@patch("chitra.cli.builder.file_check")
def test_app(mock_file_check, mock_subprocess):
    mock_file_check.return_value = True
    mock_subprocess.run = MagicMock()
    mock_subprocess.run.return_value = MagicMock()
    mock_subprocess.run.return_value.returncode = True

    app = typer.Typer()
    app.command()(builder.create)
    result = runner.invoke(app, input="Y\n")
    assert result.exit_code == 0


def test_file_check():
    with pytest.raises(UserWarning):
        file_check(["requirements.txt"])

    with pytest.raises(UserWarning):
        file_check([])



from transformers import AutoModelForSequenceClassification, AutoTokenizer, pipeline

from chitra.serve import create_api

tokenizer = AutoTokenizer.from_pretrained("microsoft/xtremedistil-l6-h256-uncased")
model = AutoModelForSequenceClassification.from_pretrained(
    "microsoft/xtremedistil-l6-h256-uncased"
)
classifier = pipeline("text-classification", model=model, tokenizer=tokenizer)

api = create_api(classifier, run=False, api_type="text-classification")

app = api.app



from chitra.utility.tf_utils import disable_gpu, gpu_dynamic_mem_growth, limit_gpu


def test_disable_gpu():
    assert disable_gpu() is None


def test_limit_gpu():
    assert limit_gpu(1, 1024) is None


def test_gpu_dynamic_mem_growth():
    assert gpu_dynamic_mem_growth() is None



from chitra.constants import _FLAX, _JAX, _TF, _TF_GPU, _TORCH, _TORCHVISION
from chitra.import_utils import is_installed

INSTALLED_MODULES = {
    module: is_installed(module)
    for module in (_TF, _TF_GPU, _TORCH, _TORCHVISION, _JAX, _FLAX)
}

_FASTAPI_INSTALLED = is_installed("fastapi")
_UVICORN_INSTALLED = is_installed("uvicorn")
_PYDANTIC_INSTALLED = is_installed("pydantic")
_MULTIPART_INSTALLED = is_installed("multipart")
_SERVE_INSTALLED = (
    _FASTAPI_INSTALLED
    and _UVICORN_INSTALLED
    and _PYDANTIC_INSTALLED
    and _MULTIPART_INSTALLED
)

_LOGURU_INSTALLED = is_installed("loguru")
_RICH_INSTALLED = is_installed("rich")



import typer
from typer.testing import CliRunner

from chitra.cli.main import version

runner = CliRunner()


def test_app():
    app = typer.Typer()
    app.command()(version)
    result = runner.invoke(app)
    assert result.exit_code == 0
    assert "You're running chitra version" in result.stdout



from transformers import AutoModelForSequenceClassification, AutoTokenizer, pipeline

from chitra.serve import create_api

tokenizer = AutoTokenizer.from_pretrained("finiteautomata/beto-sentiment-analysis")
model = AutoModelForSequenceClassification.from_pretrained(
    "finiteautomata/beto-sentiment-analysis"
)
classifier = pipeline("sentiment-analysis", model=model, tokenizer=tokenizer)

create_api(classifier, run=True, api_type="text-classification")



import io

import numpy as np
import torch
from timm import create_model

from chitra.core import load_imagenet_labels
from chitra.image import Chitra
from chitra.serve.cloud.aws_serverless import ChaliceServer

MODEL_PATH = "s3://gflow-models/efficientnet_b0.pth"

LABELS = load_imagenet_labels()


def preprocess(data) -> torch.Tensor:
    image = Chitra(data)
    image.resize((256, 256))
    x = image.numpy().astype(np.float32)
    x = x / 255.0
    x = torch.from_numpy(x)
    x = x.permute(2, 0, 1).unsqueeze(0)
    return x


def postprocess(data: torch.Tensor) -> str:
    return LABELS[data.argmax(1)]


def model_loader(buffer: io.BytesIO) -> torch.nn.Module:
    model: torch.nn.Module = create_model("efficientnet_b0", pretrained=False)
    model.load_state_dict(torch.load(buffer))
    return model


server = ChaliceServer(
    "image-classification",
    MODEL_PATH,
    model_loader=model_loader,
    preprocess_fn=preprocess,
    postprocess_fn=postprocess,
)
app = server.app
server.run("route")



from unittest.mock import MagicMock

import numpy as np
import pytest
import tensorflow as tf
from tensorflow import keras

from chitra.trainer import Dataset, InterpretModel, Trainer, create_cnn

dataset = Dataset("./")
cnn = create_cnn("mobilenetv2", num_classes=1000, keras_applications=False)
trainer = Trainer(dataset, cnn)
model_interpret = InterpretModel(True, trainer)
image_tensor = tf.random.normal((24, 24, 1))


def test_create_cnn():
    assert cnn.trainable is True
    assert isinstance(cnn, keras.models.Model)


def test_trainer():
    assert isinstance(trainer.model, keras.models.Model)


def test_interpret_model():
    assert isinstance(model_interpret.learner, Trainer)


def test_build():
    with pytest.raises(NotImplementedError):
        trainer.build()


def test_warmup():
    with pytest.raises(NotImplementedError):
        trainer.warmup()


def test_prewhiten():
    rescaled_img = trainer.prewhiten(image_tensor).numpy()
    restored_img = (rescaled_img + 1) * 127.5

    assert np.allclose(restored_img, image_tensor.numpy(), 1e-3, 1e-3)


def test_rescale():
    assert trainer.rescale(image_tensor, 0)[1] == 0


def test_summary():
    trainer.model.summary = MagicMock()
    trainer.summary()
    trainer.model.summary.assert_called_once()


def test_fit():
    trainer.model.fit = MagicMock()
    train_ds = MagicMock()
    trainer.fit(train_ds)
    trainer.model.fit.assert_called_once()


def test__get_optimizer():
    optim = trainer._get_optimizer(tf.optimizers.Adam)()
    assert isinstance(optim, tf.optimizers.Adam)



import pytest

from chitra.serve import ModelServer
from chitra.serve import constants as const


def dummy_model(x: str):
    return "hello" + x


def test_get_available_api_types():
    model_server = ModelServer.get_available_api_types()
    assert isinstance(model_server, list)


def test_model_server_run():
    model_server = ModelServer(api_type=const.TXT_CLF, model=dummy_model)
    with pytest.raises(NotImplementedError):
        model_server.run()



from typing import Callable, Dict, Optional

import uvicorn
from fastapi import FastAPI, File, UploadFile

from chitra.__about__ import documentation_url
from chitra.serve import schema
from chitra.serve.base import ModelServer
from chitra.serve.constants import IMAGE_CLF, OBJECT_DETECTION, QNA, TXT_CLF


class API(ModelServer):
    def __init__(
        self,
        api_type: str,
        model: Callable,
        preprocess_fn: Optional[Callable] = None,
        preprocess_conf: Optional[Dict] = None,
        postprocess_fn: Optional[Callable] = None,
        postprocess_conf: Optional[Dict] = None,
        **kwargs,
    ):
        """
        Creates FastAPI app for `api_type`
        Args:
            api_type: Type of the API. See `API.available_api_types()`
            model: Any ML/DL model
            preprocess_fn: Override Data Preprocessing Function, data will
            be processed with this function
            before calling model.
            postprocess_fn: Override Data Postprocessing Function, model
            output will be passed into this function.
            **kwargs:
        """
        super(API, self).__init__(
            api_type, model, preprocess_fn, postprocess_fn, **kwargs
        )

        docs_url = kwargs.get("docs_url", "/docs")
        title = kwargs.get("title", "Chitra Model Server 🔥")
        desc = kwargs.get(
            "description",
            f"<a href={documentation_url}>Goto Chitra Docs</a> 🔗",
        )

        self.app: FastAPI = FastAPI(title=title, description=desc, docs_url=docs_url)
        if not preprocess_conf:
            preprocess_conf = {}
        if not postprocess_conf:
            postprocess_conf = {}
        self.preprocess_conf = preprocess_conf
        self.postprocess_conf = postprocess_conf
        self.setup(**kwargs)

    async def predict_image(self, file: UploadFile = File(...)):
        preprocess_fn = self.data_processor.preprocess_fn
        postprocess_fn = self.data_processor.postprocess_fn

        x = preprocess_fn(await file.read())
        x = self.model(x)
        x = postprocess_fn(x)
        return x

    async def predict_text(self, data: schema.Query):
        data_processor = self.data_processor
        x = data.query
        if data_processor.preprocess_fn:
            x = data_processor.preprocess(x)
        x = self.model(x)
        if data_processor.postprocess_fn:
            x = data_processor.postprocess(x)
        return x

    async def predict_question_answer(self, data: schema.QnARequest):
        data_processor = self.data_processor
        x = data.query, data.question
        if data_processor.preprocess_fn:
            x = data_processor.preprocess(x)
        x = self.model(x)
        if data_processor.postprocess_fn:
            x = data_processor.postprocess(x)
        return x

    def setup(self, **_):

        if self.api_type in (IMAGE_CLF, OBJECT_DETECTION):
            self.app.post("/api/predict-image")(self.predict_image)

        elif self.api_type == TXT_CLF:
            self.app.post("/api/predict-text")(self.predict_text)

        elif self.api_type == QNA:
            self.app.post("/api/QnA")(self.predict_question_answer)

    def run(self):
        uvicorn.run(self.app)


def create_api(
    model: Callable,
    api_type: str = "IMAGE-CLASSIFICATION",
    preprocess_fn: Callable = None,
    preprocess_conf: Optional[Dict] = None,
    postprocess_fn: Callable = None,
    postprocess_conf: Optional[Dict] = None,
    run: bool = False,
    **kwargs,
) -> API:
    """
    Launch FastAPI app
    Args:
        model: Any ML/DL model
        api_type: Type of the API task, see `chitra.serve.get_available_api_types()`
        preprocess_fn: Override default preprocessing function
        preprocess_conf: Arguments for preprocessing function
        postprocess_fn: Override default postprocessing function
        postprocess_conf: Arguments for postprocessing function
        run: Set True to run the app
        **kwargs:

    Returns:
        Object of `chitra.serve.API` class
    """
    api = API(
        api_type,
        model,
        preprocess_fn=preprocess_fn,
        preprocess_conf=preprocess_conf,
        postprocess_fn=postprocess_fn,
        postprocess_conf=postprocess_conf,
        **kwargs,
    )

    if run:
        api.run()
    return api



import os

import tensorflow as tf

from chitra.image import Chitra
from chitra.tf_image import read_image, resize_image

chitra_banner = (
    "https://raw.githubusercontent.com/aniketmaurya/"
    "chitra/master/docs/assets/chitra_banner.png"
)
image = Chitra(chitra_banner).image


def test_read_image():
    image_file = "./test_read_image.png"
    image.save(image_file)
    tf_image = read_image(image_file)
    assert 3 <= tf_image.shape[-1] <= 4
    assert isinstance(tf_image, tf.Tensor)
    os.remove(image_file)


def test_resize_image():
    assert resize_image(tf.random.normal((32, 32, 3)), (24, 24)).shape == (24, 24, 3)



from pydantic import BaseModel


class Query(BaseModel):
    query: str


class QueryResult(Query):
    result: str


class QnARequest(Query):
    question: str


class QnAResponse(QnARequest):
    result: str



import itertools
from typing import Callable, List, Optional

from chitra.data_processing import (
    DataProcessor,
    DefaultTextProcessor,
    DefaultVisionProcessor,
)
from chitra.serve import constants as const


class ModelServer:
    API_TYPES = {
        "VISION": (const.IMAGE_CLF, const.OBJECT_DETECTION),
        "NLP": (const.TXT_CLF, const.QNA),
    }

    def __init__(
        self,
        api_type: str,
        model: Callable,
        preprocess_fn=None,
        postprocess_fn=None,
        preprocess_conf: Optional[dict] = None,
        postprocess_conf: Optional[dict] = None,
        **kwargs,
    ):
        if not preprocess_conf:
            preprocess_conf = {}
        if not postprocess_conf:
            postprocess_conf = {}

        self.api_type = api_type.upper()
        self.model = model
        self.preprocess_conf = preprocess_conf
        self.postprocess_conf = postprocess_conf
        self.data_processor: Optional[DataProcessor] = self.set_data_processor(
            preprocess_fn, postprocess_fn
        )

    @classmethod
    def get_available_api_types(cls) -> List[str]:
        return list(itertools.chain.from_iterable(cls.API_TYPES.values()))

    def set_data_processor(
        self, preprocess_fn: Callable, postprocess_fn: Callable
    ) -> DataProcessor:
        data_preprocessor = self.set_default_processor()
        if preprocess_fn:
            data_preprocessor.set_preprocess_fn(preprocess_fn)
        if postprocess_fn:
            data_preprocessor.set_postprocess_fn(postprocess_fn)
        return data_preprocessor

    def set_default_processor(self) -> DataProcessor:
        api_type = self.api_type
        if api_type in ModelServer.API_TYPES.get("VISION"):
            self.data_processor = DefaultVisionProcessor.vision
        elif api_type in ModelServer.API_TYPES.get("NLP"):
            self.data_processor = DefaultTextProcessor.nlp
        else:
            raise UserWarning(
                f"{api_type} is not implemented! Available types are -\
                 {ModelServer.get_available_api_types()}"
            )
        return self.data_processor

    def run(self, *_, **__):
        raise NotImplementedError



import os

import tensorflow as tf

from chitra.core import get_basename, load_imagenet_labels, remove_dsstore


def test_remove_dsstore():
    os.makedirs("chitra_temp", exist_ok=True)
    ds_store = "chitra_temp/.DS_Store"
    open(ds_store, "w").close()
    assert os.path.exists(ds_store)
    remove_dsstore("chitra_temp")
    assert not os.path.exists(ds_store)
    os.removedirs("chitra_temp")


def test_get_basename():
    assert get_basename(tf.constant("hello/world")) == "world"


def test_load_imagenet_labels():
    labels = load_imagenet_labels()

    assert "\n" not in labels
    assert len(labels) == 1000 + 1



import unittest.mock
from unittest.mock import MagicMock

import numpy as np
import pytest

from chitra.serve import GradioApp
from chitra.serve import constants as const


def dummy_model(x):
    return np.random.randn(1)


def preprocess_fn(x, rescale: bool, expand_dims: bool):
    if rescale:
        x = x / 127.5 - 1.0
    if expand_dims:
        x = np.expand_dims(x, 0)
    return x


def postprocess_fn(x, thresh: float):
    return (x > thresh)[0]


def test_gradio_app():
    api_types = GradioApp.get_available_api_types()
    for api_type in api_types:
        app = GradioApp(api_type, model=dummy_model)
        assert app

    for api_type in api_types:
        app = GradioApp(
            api_type,
            model=dummy_model,
            preprocess_fn=lambda dummy: dummy,
            postprocess_fn=lambda dummy: dummy,
        )
        assert app


def test_image_classification():
    dummy_image = np.random.randn(224, 224, 3)
    preprocess_conf = {"rescale": True, "expand_dims": True}
    postprocess_conf = {"thresh": 0.5}

    app = GradioApp(
        const.IMAGE_CLF,
        model=dummy_model,
        preprocess_fn=preprocess_fn,
        preprocess_conf=preprocess_conf,
        postprocess_fn=postprocess_fn,
        postprocess_conf=postprocess_conf,
    )

    assert app.single_x_classification(dummy_image) in (0, 1)


@pytest.mark.parametrize(
    "test_input, expected",
    [(None, {}), ({"article": "chitra testing"}, {"article": "chitra testing"})],
)
@unittest.mock.patch("chitra.serve.app.gr")
def test_run(mock_gr, test_input, expected):
    mock_gr.Interface = MagicMock()
    app = GradioApp(const.IMAGE_CLF, model=dummy_model)
    app.run(gr_interface_conf=test_input)

    mock_gr.Interface.assert_called_with(
        fn=app.api_type_func[app.api_type],
        inputs=app.input_types,
        outputs=app.output_types,
        title=app.title,
        description=app.desc,
        **expected,
    )



import numpy as np

from chitra.data_processing.default.vision import default_preprocess


def test_default_preprocess():
    image_arr = np.random.randn(16, 16, 3).astype("uint8")

    result = default_preprocess(
        data=image_arr, image_shape=(32, 32), rescale=True, expand_dims=True
    )
    assert isinstance(result, np.ndarray)
    assert result.shape == (1, 32, 32, 3)



from chitra.import_utils import is_installed

onnx = None
tf2onnx = None
onnx2pytorch = None
torch = None

if is_installed("onnx"):
    import onnx

if is_installed("tf2onnx"):
    import tf2onnx

if is_installed("torch"):
    import torch.onnx


def pytorch_to_onnx(model, tensor, export_path="temp.onnx"):
    # Input to the model
    torch_out = model(tensor)

    # Export the model
    torch.onnx.export(
        model,  # model being run
        tensor,  # model input (or a tuple for multiple inputs)
        export_path,  # where to save the model (can be a file or file-like object)
        export_params=True,  # store the trained parameter weights inside the model file
        opset_version=10,  # the ONNX version to export the model to
        do_constant_folding=True,  # whether to execute constant folding for optimization
        input_names=["input"],  # the model's input names
        output_names=["output"],  # the model's output names
        dynamic_axes={
            "input": {0: "batch_size"},  # variable length axes
            "output": {0: "batch_size"},
        },
    )
    return export_path


def tf2_to_onnx(model, opset=None, output_path=None, **kwargs):
    inputs_as_nchw = kwargs.get("inputs_as_nchw", "input0:0")
    onnx_model = tf2onnx.convert.from_keras(
        model, opset=opset, output_path=output_path, inputs_as_nchw=inputs_as_nchw
    )
    return onnx_model



from chitra.trainer import Dataset

dataset = Dataset("./")


def test__process():
    assert True


def test__reload():
    assert True


def test__capture_return_types():
    assert True


def test_update_component():
    assert True


def test_get_labels():
    assert True


def test_label_encoder():
    assert True


def test_generator():
    assert True


def test_get_tf_dataset():
    assert True



import io
import os
from io import BytesIO
from pathlib import Path
from typing import Any, List, Union

import matplotlib.pyplot as plt
import numpy as np
import requests
from PIL import Image

from chitra.constants import _TF, _TORCH, CHITRA_URL_SEP, IMAGE_CACHE_DIR
from chitra.coordinates import BoundingBoxes
from chitra.import_utils import is_installed

tf = None
torch = None

if is_installed(_TF):
    import tensorflow as tf

if is_installed(_TORCH):
    import torch

DATA_FORMATS = Union[str, Image.Image, np.ndarray, "tf.Tensor", "torch.Tensor"]
DEFAULT_MODE = os.environ.get("CHITRA_DEFAULT_MODE", "TF")


def _cache_image(image: Image.Image, image_path: str):
    cache_dir = Path(IMAGE_CACHE_DIR)
    filename = image_path.replace("/", CHITRA_URL_SEP)
    os.makedirs(cache_dir, exist_ok=True)
    image.save(cache_dir / filename)


def _url_to_image(url: str, cache: bool) -> Image.Image:
    """returns Image from url."""
    filename = url.replace("/", CHITRA_URL_SEP)
    cache_file = Path(IMAGE_CACHE_DIR) / filename
    if cache and os.path.exists(cache_file):
        return Image.open(cache_file)

    if not url.lower().startswith("http"):
        raise AssertionError("invalid url, must start with http")
    content = requests.get(url).content
    image = Image.open(BytesIO(content))
    if cache:
        _cache_image(image, url)
    return image


class Chitra:
    """Ultimate image utility class.

    1. Load image from file, web url, numpy or bytes
    2. Plot image
    3. Draw bounding boxes
    """

    def __init__(
        self,
        data: Any,
        bboxes: List = None,
        labels: List = None,
        box_format: str = BoundingBoxes.CORNER,
        cache: bool = False,
        *args,
        **kwargs
    ) -> None:
        """

        Args:
            data: numpy, url, filelike
            bboxes:
            labels:
            box_format:
            cache[bool]: Whether to cache downloaded image
            *args:
            **kwargs:
        """
        super().__init__()
        self.image = self._load_image(data, cache=cache)
        self.bboxes = None

        if bboxes is not None:
            self.bboxes = BoundingBoxes(bboxes, labels)

    @staticmethod
    def _load_image(data: DATA_FORMATS, cache: bool):
        if isinstance(data, Image.Image):
            return data

        elif isinstance(data, bytes):
            return Image.open(io.BytesIO(data))

        if isinstance(data, (tf.Tensor, torch.Tensor)):
            data = data.numpy().astype("uint8")

        if isinstance(data, str):
            if data.startswith("http"):
                image = _url_to_image(data, cache)

            else:
                image = Image.open(data)

        elif isinstance(data, np.ndarray):
            image = Image.fromarray(data)

        else:
            raise UserWarning("unable to load image!")

        return image

    def numpy(self):
        return np.asarray(self.image)

    def to_tensor(self, mode: str = DEFAULT_MODE):
        """mode: tf/torch/pt"""
        mode = mode.upper()
        np_image = self.numpy()

        if mode == "TF":
            tensor = tf.constant(np_image)
        elif mode in ("TORCH", "PT"):
            tensor = torch.from_numpy(np_image)
        else:
            raise UserWarning("invalid mode!")
        return tensor

    @property
    def shape(self):
        return self.numpy().shape

    @property
    def size(self):
        return self.image.size

    def imshow(self, cmap=plt.cm.Blues, *args, **kwargs):
        plt.imshow(self.numpy(), cmap, *args, **kwargs)

    def draw_boxes(
        self,
        marker_size: int = 2,
        color=(0, 255, 0),
    ):
        if self.bboxes is None:
            raise UserWarning("bboxes is None")

        bbox_on_image = self.bboxes.get_bounding_boxes_on_image(self.shape)
        return bbox_on_image.draw_on_image(
            self.numpy()[..., :3], color=color, size=marker_size
        )

    def resize(self, *args, **kwargs) -> Image.Image:
        """
        Calls PIL.Image.resize method and passes the arguments
        Args:
            *args:
            **kwargs:

        Returns:
            resized PIL.Image
        """
        self.image = self.image.resize(*args, **kwargs)
        return self.image

    def resize_image_with_bbox(self, size: List[int]):
        old_size = self.shape
        self.image = self.image.resize(size)
        self.bboxes.resize_with_image(old_size, self.numpy())
        return self.image, self.bboxes



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['chitra',
 'chitra.cli',
 'chitra.converter',
 'chitra.data_processing',
 'chitra.data_processing.default',
 'chitra.serve',
 'chitra.serve.cloud',
 'chitra.serve.schema',
 'chitra.serve.tf_serving',
 'chitra.serve.torchserve',
 'chitra.utility',
 'chitra.visualization']

package_data = \
{'': ['*'], 'chitra': ['assets/*']}

install_requires = \
['tensorflow >= 2.3',
 'tensorflow-addons >=0.13.0',
 'tf-keras-vis >=0.5.3',
 'pytorch-lightning',
 'matplotlib',
 'pillow',
 'imgaug >=0.4.0',
 'scikit-learn',
 'typeguard',
 'gradio >=2.2.2',
 'typer']

extras_require = \
{'converter': ['onnx', 'tf2onnx'],
 'serve': ['fastapi',
           'uvicorn',
           'pydantic',
           'python-multipart',
           'tensorflow-serving-api',
           'chalice',
           'smart_open'],
 'test': ['pytest', 'pytest-asyncio', 'coverage']}

entry_points = \
{'console_scripts': ['chitra = chitra.cli.main:app']}

setup(name='chitra',
      version='0.2.0a1',
      description='Deep Learning library for Model Building, Interpretability, Visualization, API Building & Deployment.',
      author='Aniket Maurya',
      author_email='hello@aniketmaurya.com',
      url='https://github.com/aniketmaurya/chitra',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.7',
     )



import os
import pathlib
from typing import Optional, Tuple

import requests
import tensorflow as tf

from chitra.constants import IMAGENET_LABEL_URL
from chitra.logging import logger

IMAGENET_LABELS: Optional[Tuple[str]] = None


def remove_dsstore(path) -> None:
    """Deletes .DS_Store files from path and sub-folders of path."""
    path = pathlib.Path(path)

    for e in path.glob("*.DS_Store"):
        os.remove(e)

    for e in path.glob("*/*.DS_Store"):
        os.remove(e)


def get_basename(path: tf.string) -> tf.string:
    if not isinstance(path, tf.Tensor):
        raise AssertionError
    return tf.strings.split(path, os.path.sep)[-1]


def load_imagenet_labels() -> Tuple[str]:
    global IMAGENET_LABELS
    if IMAGENET_LABELS is None:
        logger.debug("Downloading imagenet labels...")
        IMAGENET_LABELS = (
            requests.get(IMAGENET_LABEL_URL).content.decode("UTF-8").split("\n")[1:]
        )
    return IMAGENET_LABELS



import io
from abc import ABC
from typing import Callable

import smart_open

from chitra.serve.base import ModelServer


class CloudServer(ModelServer, ABC):
    def __init__(
        self,
        api_type: str,
        model_path: str,
        model_loader: Callable,
        preprocess_fn: Callable,
        postprocess_fn: Callable,
        **kwargs
    ):
        raw_model = self.download_model(model_path, **kwargs)
        model = model_loader(raw_model)

        super().__init__(
            api_type,
            model,
            preprocess_fn=preprocess_fn,
            postprocess_fn=postprocess_fn,
            **kwargs
        )

    def download_model(self, path: str, **kwargs) -> io.BytesIO:
        """
        Download model from cloud
        ref: http://5.9.10.113/67706477/load-pytorch-model-from-s3-bucket
        Args:
            path:
            **kwargs:

        Returns:

        """

        with smart_open.open(path, mode="rb", **kwargs) as fr:
            data = io.BytesIO(fr.read())
        return data

    def run(self, *_, **__):
        raise NotImplementedError



from typing import Callable, List, Optional

from chalice import Chalice, Rate
from loguru import logger

from chitra.serve.cloud.base import CloudServer

S3 = "s3"
GCS = "gcs"

RATE_UNIT = {"m": Rate.MINUTES, "h": Rate.HOURS, "d": Rate.DAYS}


def infer_location_type(path: str):
    if path.startswith("s3"):
        return S3
    if path.startswith("gcs"):
        return GCS
    raise ValueError(f"Location type is not supported yet for path={path}")


class ChaliceServer(CloudServer):
    INVOKE_METHODS = ("route",)

    def __init__(
        self,
        api_type: str,
        model_path: str,
        model_loader: Callable,
        preprocess_fn: Callable = None,
        postprocess_fn: Callable = None,
        **kwargs,
    ):
        super().__init__(
            api_type,
            model_path=model_path,
            model_loader=model_loader,
            preprocess_fn=preprocess_fn,
            postprocess_fn=postprocess_fn,
            **kwargs,
        )

        self.app = Chalice(app_name=kwargs.get("name", "chitra-server"))

    def index(self):
        return {"hello": "world"}

    def predict(self) -> dict:

        data_processor = self.data_processor
        x = self.app.current_request.raw_body
        logger.debug(f"raw body type={type(x)}")
        logger.debug(x)
        if data_processor.preprocess_fn:
            x = data_processor.preprocess(x, **self.preprocess_conf)
        x = self.model(x)
        if data_processor.postprocess_fn:
            x = data_processor.postprocess(x, **self.postprocess_conf)
        return x

    def run(self, invoke_method: str, content_types: Optional[List] = None, **kwargs):
        invoke_method = invoke_method.lower()
        if not content_types:
            content_types = []

        if invoke_method not in self.INVOKE_METHODS:
            raise NotImplementedError(
                f"invoke method={invoke_method} not implemented yet. Please select {self.INVOKE_METHODS}"
            )

        if invoke_method == "route":
            route_path = kwargs.get("path", "/predict")
            self.app.route("/", methods=["GET"])(self.index)
            self.app.route(route_path, methods=["POST"], content_types=content_types)(
                self.predict
            )



from typing import Callable, Optional


class DataProcessor:
    def __init__(
        self,
        preprocess_fn: Optional[Callable] = None,
        postprocess_fn: Optional[Callable] = None,
    ):
        self._preprocess_fn = preprocess_fn
        self._postprocess_fn = postprocess_fn

    @property
    def preprocess_fn(self):
        return self._preprocess_fn

    @property
    def postprocess_fn(self):
        return self._postprocess_fn

    def set_preprocess_fn(self, func):
        self._preprocess_fn = func

    def set_postprocess_fn(self, func):
        self._postprocess_fn = func

    def preprocess(self, x, **kwargs):
        if self._preprocess_fn is None:
            raise UserWarning("preprocess method is not defined")
        return self._preprocess_fn(x, **kwargs)

    def postprocess(self, x, **kwargs):
        if self._postprocess_fn is None:
            raise UserWarning("postprocess method not defined")
        return self._postprocess_fn(x, **kwargs)
